import pytest
from mock import Mock, patch, call

from ramses import generators
from .fixtures import engine_mock


class TestGenerationHelpers(object):

    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_existing_model(self, mock_get):
        mock_get.return_value = 1
        model, auth_model = generators.setup_data_model('foo', 'Bar')
        assert not auth_model
        assert model == 1
        mock_get.assert_called_once_with('Bar')

    @patch('ramses.generators.resource_schema')
    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_no_schema(self, mock_get, mock_schema):
        mock_get.return_value = None
        mock_schema.return_value = None
        with pytest.raises(Exception) as ex:
            generators.setup_data_model('foo', 'Bar')
        assert str(ex.value) == 'Missing schema for model `Bar`'
        mock_get.assert_called_once_with('Bar')
        mock_schema.assert_called_once_with('foo')

    @patch('ramses.generators.resource_schema')
    @patch('ramses.models.generate_model_cls')
    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_success(self, mock_get, mock_gen, mock_schema):
        mock_get.return_value = None
        mock_schema.return_value = {'field1': 'val1'}
        model = generators.setup_data_model('foo', 'Bar')
        mock_get.assert_called_once_with('Bar')
        mock_schema.assert_called_once_with('foo')
        mock_gen.assert_called_once_with(
            schema={'field1': 'val1'},
            model_name='Bar',
            raml_resource='foo')
        assert model == mock_gen()

    @patch('ramses.generators.setup_data_model')
    def test_handle_model_generation_value_err(self, mock_set):
        mock_set.side_effect = ValueError('strange error')
        with pytest.raises(ValueError) as ex:
            generators.handle_model_generation('foo', '/stories')
        assert str(ex.value) == 'Story: strange error'
        mock_set.assert_called_once_with('foo', 'Story')

    @patch('ramses.generators.setup_data_model')
    def test_handle_model_generation(self, mock_set):
        mock_set.return_value = ('Foo1', True)
        model, auth_model = generators.handle_model_generation(
            'foo', '/stories')
        mock_set.assert_called_once_with('foo', 'Story')
        assert model == 'Foo1'
        assert auth_model

    @patch('ramses.generators.configure_resources')
    def test_generate_server(self, mock_conf):
        parsed_raml = Mock(resources={'/users': 1})
        config = Mock()
        generators.generate_server(parsed_raml, config)
        mock_conf.assert_called_once_with(
            config=config, raml_resources={'/users': 1},
            parsed_raml=parsed_raml)


class TestGenerateModels(object):

    @patch('ramses.generators.is_dynamic_uri')
    def test_no_resources(self, mock_dyn):
        generators.generate_models(config=1, raml_resources={})
        assert not mock_dyn.called

    @patch('ramses.generators.handle_model_generation')
    def test_dynamic_uri(self, mock_handle):
        generators.generate_models(
            config=1, raml_resources={'/{id}': Mock(resources={})})
        assert not mock_handle.called

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.generators.handle_model_generation')
    def test_attr_subresource(self, mock_handle, mock_attr):
        mock_attr.return_value = True
        resource = Mock(resources={})
        generators.generate_models(
            config=1, raml_resources={'/stories': resource})
        assert not mock_handle.called
        mock_attr.assert_called_once_with(resource, 'stories')

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.generators.handle_model_generation')
    def test_non_auth_model(self, mock_handle, mock_attr):
        mock_attr.return_value = False
        mock_handle.return_value = ('Foo', False)
        config = Mock()
        resource = Mock(resources={})
        generators.generate_models(
            config=config, raml_resources={'/stories': resource})
        mock_attr.assert_called_once_with(resource, 'stories')
        mock_handle.assert_called_once_with(resource, 'stories')
        assert config.registry.auth_model != 'Foo'

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.generators.handle_model_generation')
    def test_auth_model(self, mock_handle, mock_attr):
        mock_attr.return_value = False
        mock_handle.return_value = ('Foo', True)
        config = Mock()
        resource = Mock(resources={})
        generators.generate_models(
            config=config, raml_resources={'/stories': resource})
        mock_attr.assert_called_once_with(resource, 'stories')
        mock_handle.assert_called_once_with(resource, 'stories')
        assert config.registry.auth_model == 'Foo'

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.generators.handle_model_generation')
    def test_recursion(self, mock_handle, mock_attr):
        mock_attr.return_value = False
        mock_handle.return_value = ('Foo', False)
        resource1 = Mock(resources={})
        resource2 = Mock(resources={'/users': resource1})
        generators.generate_models(
            config='', raml_resources={'/stories': resource2})
        mock_handle.assert_has_calls([
            call(resource2, 'stories'),
            call(resource1, 'users'),
        ])


class TestConfigureResources(object):

    @patch('ramses.generators.is_restful_uri')
    def test_no_raml_resources(self, mock_rest):
        config = Mock()
        generators.configure_resources(
            config, raml_resources={}, parsed_raml='',
            parent_resource=None)
        assert not config.get_root_resource.called
        assert not mock_rest.called

    def test_no_parent_not_restful_uri(self):
        config = Mock()
        with pytest.raises(ValueError) as ex:
            generators.configure_resources(
                config, raml_resources={'/foo/bar': ''},
                parsed_raml='', parent_resource=None)
        expected = 'Resource URI `/foo/bar` is not RESTful'
        assert str(ex.value) == expected
        config.get_root_resource.assert_called_once_with()

    @patch('ramses.generators.singular_subresource')
    def test_root_dynamic_resource(self, mock_singular):
        config = Mock()
        resource = Mock(resource={})
        with pytest.raises(Exception) as ex:
            generators.configure_resources(
                config, raml_resources={'/{id}': resource},
                parsed_raml='', parent_resource=None)
        assert "Top-level resources can't be dynamic" in str(ex.value)
        assert not mock_singular.called

    @patch('ramses.generators.singular_subresource')
    def test_dynamic_resource(self, mock_singular):
        resource = Mock(resources={})
        parent_resource = Mock(uid=1)
        generators.configure_resources(
            None, raml_resources={'/{id}': resource},
            parsed_raml='', parent_resource=parent_resource)
        assert not mock_singular.called

    @patch('ramses.generators.singular_subresource')
    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.get_existing_model')
    @patch('ramses.generators.generate_acl')
    @patch('ramses.generators.resource_view_attrs')
    @patch('ramses.generators.generate_rest_view')
    def test_full_run(
            self, generate_view, view_attrs, generate_acl, get_model,
            attr_res, singular_res):
        model_cls = Mock()
        model_cls.pk_field.return_value = 'my_id'
        attr_res.return_value = False
        singular_res.return_value = False
        get_model.return_value = model_cls
        config = Mock()
        resource = Mock(resources={})
        parent_resource = Mock(uid=1)

        generators.configure_resources(
            config, raml_resources={'/stories': resource},
            parsed_raml='foo', parent_resource=parent_resource)

        singular_res.assert_called_once_with(resource, 'stories')
        attr_res.assert_called_once_with(resource, 'stories')
        get_model.assert_called_once_with('Story')
        generate_acl.assert_called_once_with(
            context_cls=model_cls,
            raml_resource=resource,
            parsed_raml='foo'
        )
        view_attrs.assert_called_once_with(resource, False)
        generate_view.assert_called_once_with(
            model_cls=model_cls,
            attrs=view_attrs(),
            attr_view=False,
            singular=False
        )
        parent_resource.add.assert_called_once_with(
            'story', 'stories',
            id_name='stories_my_id',
            factory=generate_acl(),
            view=generate_view()
        )

    @patch('ramses.generators.singular_subresource')
    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.get_existing_model')
    @patch('ramses.generators.generate_acl')
    @patch('ramses.generators.resource_view_attrs')
    @patch('ramses.generators.generate_rest_view')
    def test_full_run_singular(
            self, generate_view, view_attrs, generate_acl, get_model,
            attr_res, singular_res):
        attr_res.return_value = False
        singular_res.return_value = True
        config = Mock()
        resource = Mock(resources={})
        parent_resource = Mock(uid=1)
        parent_resource.view._model_class.pk_field.return_value = 'other_id'

        generators.configure_resources(
            config, raml_resources={'/stories': resource},
            parsed_raml='foo', parent_resource=parent_resource)

        singular_res.assert_called_once_with(resource, 'stories')
        attr_res.assert_called_once_with(resource, 'stories')
        get_model.assert_called_once_with('Story')
        generate_acl.assert_called_once_with(
            context_cls=parent_resource.view._model_class,
            raml_resource=resource,
            parsed_raml='foo'
        )
        view_attrs.assert_called_once_with(resource, True)
        generate_view.assert_called_once_with(
            model_cls=parent_resource.view._model_class,
            attrs=view_attrs(),
            attr_view=False,
            singular=True
        )
        parent_resource.add.assert_called_once_with(
            'story',
            factory=generate_acl(),
            view=generate_view()
        )
        assert generate_view()._singular_model == get_model()



from __future__ import print_function

import logging

from inflection import pluralize, singularize

from .views import generate_rest_view
from .acl import generate_acl
from .utils import (
    is_dynamic_uri, resource_view_attrs, generate_model_name,
    is_restful_uri, dynamic_part_name, resource_schema,
    attr_subresource, singular_subresource)


log = logging.getLogger(__name__)


def setup_data_model(raml_resource, model_name):
    """ Setup storage/data model and return generated model class.

    Process follows these steps:
      * Resource schema is found and restructured by `resource_schema`.
      * Model class is generated from properties dict using util function
        `generate_model_cls`.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource.
        :model_name: String representing model name.
    """
    from .models import generate_model_cls, get_existing_model
    model_cls = get_existing_model(model_name)
    if model_cls is not None:
        return model_cls, False

    schema = resource_schema(raml_resource)
    if not schema:
        raise Exception('Missing schema for model `{}`'.format(model_name))

    log.info('Generating model class `{}`'.format(model_name))
    return generate_model_cls(
        schema=schema,
        model_name=model_name,
        raml_resource=raml_resource,
    )


def handle_model_generation(raml_resource, route_name):
    """ Generates model name and runs `setup_data_model` to get
    or generate actual model class.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource.
        :route_name: String name of the resource.
    """
    model_name = generate_model_name(route_name)
    try:
        return setup_data_model(raml_resource, model_name)
    except ValueError as ex:
        raise ValueError('{}: {}'.format(model_name, str(ex)))


def configure_resources(config, raml_resources, parsed_raml,
                        parent_resource=None):
    """ Perform complete resource configuration process

    RAML data from `raml_resources` is used. Created resources
    are attached to `parent_resource` class which is an instance of
    `nefertari.resource.Resource`.

    This function iterates through resources data from `raml_resources` and
    generates: ACL, view, route, resource,
    database model. It is called recursively for configuring child resources.

    Things to consider:
      * Top-level resources must be collection names.
      * Resource nesting must look like collection/id/collection/id/...
      * No resources are explicitly created for dynamic (ending with '}')
        RAML resources as they are implicitly processed by parent collection
        resources.
      * Collection resources can only have 1 dynamic child resource.

    Arguments:
        :config: Pyramid Configurator instance
        :raml_resources: Map of {uri_string: pyraml.entities.RamlResource}
        :parsed_raml: Whole parsed RAML object
        :parent_resource: Instance of `nefertari.resource.Resource`
    """
    from .models import get_existing_model
    if not raml_resources:
        return

    # Use root factory for root-level resources
    parent_arg = parent_resource
    if parent_resource is None:
        parent_resource = config.get_root_resource()

    for resource_uri, raml_resource in raml_resources.items():
        if not is_restful_uri(resource_uri):
            raise ValueError('Resource URI `{}` is not RESTful'.format(
                resource_uri))

        clean_uri = route_name = resource_uri.strip('/')
        log.info('Configuring resource: `{}`. Parent: `{}`'.format(
            route_name, parent_resource.uid or 'root'))

        # No need to setup routes/views for dynamic resource as it was already
        # setup when parent was configured.
        if is_dynamic_uri(resource_uri):
            if parent_arg is None:
                raise Exception("Top-level resources can't be dynamic and must "
                                "represent collections instead")
            return configure_resources(
                config=config,
                raml_resources=raml_resource.resources,
                parsed_raml=parsed_raml,
                parent_resource=parent_resource)

        # Generate DB model
        # If this is an attribute or singular resource, we don't need to
        # generate model
        is_singular = singular_subresource(raml_resource, route_name)
        is_attr_res = attr_subresource(raml_resource, route_name)
        if parent_arg is not None and (is_attr_res or is_singular):
            model_cls = parent_resource.view._model_class
        else:
            model_name = generate_model_name(route_name)
            model_cls = get_existing_model(model_name)

        resource_kwargs = {}

        # Generate ACL. Use GuestACL for now
        log.info('Generating ACL for `{}`'.format(route_name))
        resource_kwargs['factory'] = generate_acl(
            context_cls=model_cls,
            raml_resource=raml_resource,
            parsed_raml=parsed_raml,
        )

        # Generate dynamic part name
        if not is_singular:
            resource_kwargs['id_name'] = dynamic_part_name(
                raml_resource=raml_resource,
                clean_uri=clean_uri,
                pk_field=model_cls.pk_field())

        # Generate REST view
        log.info('Generating view for `{}`'.format(route_name))
        resource_kwargs['view'] = generate_rest_view(
            model_cls=model_cls,
            attrs=resource_view_attrs(raml_resource, is_singular),
            attr_view=is_attr_res,
            singular=is_singular,
        )

        # In case of singular resource, model still needs to be generated,
        # but we store it on a different view attribute
        if is_singular:
            model_name = generate_model_name(route_name)
            resource_kwargs['view']._singular_model = get_existing_model(
                model_name)

        # Create new nefertari resource
        log.info('Creating new resource for `{}`'.format(route_name))
        resource_args = (singularize(clean_uri),)

        if not is_singular:
            resource_args += (pluralize(clean_uri),)

        new_resource = parent_resource.add(*resource_args, **resource_kwargs)

        # Configure child resources if present
        configure_resources(
            config=config,
            raml_resources=raml_resource.resources,
            parsed_raml=parsed_raml,
            parent_resource=new_resource)


def generate_server(parsed_raml, config):
    """ Run server generation process.

    Arguments:
        :config: Pyramid Configurator instance.
        :parsed_raml: Parsed pyraml structure.
    """
    log.info('Server generation started')

    # Setup resources
    configure_resources(
        config=config, raml_resources=parsed_raml.resources,
        parsed_raml=parsed_raml)


def generate_models(config, raml_resources):
    """ Generate model for each resource in :raml_resources:

    Notes:
      * The DB model name is generated using singular titled version of current
        resource's url. E.g. for resource under url '/stories', model with
        name 'Story' will be generated.

    Arguments:
        :config: Pyramid Configurator instance
        :raml_resources: Map of {uri_string: pyraml.entities.RamlResource}
    """
    if not raml_resources:
        return

    for resource_uri, raml_resource in raml_resources.items():
        # No need to generate models for dynamic resource
        if is_dynamic_uri(resource_uri):
            return generate_models(
                config, raml_resources=raml_resource.resources)

        # Generate DB model
        # If this is an attribute resource we don't need to generate model
        route_name = resource_uri.strip('/')
        if not attr_subresource(raml_resource, route_name):
            log.info('Configuring model for route `{}`'.format(route_name))
            model_cls, is_auth_model = handle_model_generation(
                raml_resource, route_name)
            if is_auth_model:
                config.registry.auth_model = model_cls

        # Generate child models if present
        generate_models(config, raml_resources=raml_resource.resources)



import logging

from nefertari import engine
from nefertari.authentication.models import AuthModelDefaultMixin
from .utils import find_dynamic_resource
from . import registry


log = logging.getLogger(__name__)

"""
Map of RAML types names to nefertari.engine fields.

"""
type_fields = {
    'string':           engine.StringField,
    'float':            engine.FloatField,
    'integer':          engine.IntegerField,
    'boolean':          engine.BooleanField,
    'datetime':         engine.DateTimeField,
    'file':             engine.BinaryField,
    'relationship':     engine.Relationship,
    'dict':             engine.DictField,
    'foreign_key':      engine.ForeignKeyField,
    'big_integer':      engine.BigIntegerField,
    'date':             engine.DateField,
    'choice':           engine.ChoiceField,
    'interval':         engine.IntervalField,
    'decimal':          engine.DecimalField,
    'pickle':           engine.PickleField,
    'small_integer':    engine.SmallIntegerField,
    'text':             engine.TextField,
    'time':             engine.TimeField,
    'unicode':          engine.UnicodeField,
    'unicode_text':     engine.UnicodeTextField,
    'id_field':         engine.IdField,
    'list':             engine.ListField,
}


def get_existing_model(model_name):
    """ Try to find existing model class named `model_name`.

    Arguments:
        :model_name: String name of the model class.
    """
    try:
        model_cls = engine.get_document_cls(model_name)
        log.debug('Model `{}` already exists. Using existing one'.format(
            model_name))
        return model_cls
    except ValueError:
        log.debug('Model `{}` does not exist'.format(model_name))


def prepare_relationship(field_name, model_name, raml_resource):
    """ Create referenced model if it doesn't exist.

    When preparing a relationship, we check to see if the model that will be
    referenced already exists. If not, it is created so that it will be possible
    to use it in a relationship. Thus the first usage of this model in RAML file
    must provide its schema in one of http methods that are assumed to contain
    a full body schema.

    Arguments:
        :field_name: Name of the field that should become a `Relationship`.
        :raml_resource: Instance of pyraml.entities.RamlResource. Resource
            for which :model_name: will is being defined.
    """
    from .generators import setup_data_model
    if get_existing_model(model_name) is None:
        dynamic_res = find_dynamic_resource(raml_resource)
        subresources = getattr(dynamic_res, 'resources', {}) or {}
        subresources = {k.strip('/'): v for k, v in subresources.items()}
        if field_name not in subresources:
            raise ValueError('Model `{}` used in relationship `{}` is not '
                             'defined'.format(model_name, field_name))
        setup_data_model(subresources[field_name], model_name)


def generate_model_cls(schema, model_name, raml_resource, es_based=True):
    """ Generate model class.

    Engine DB field types are determined using `type_fields` and only those
    types may be used.

    Arguments:
        :properties: Dictionary of DB schema fields which looks like
            {field_name: {required: boolean, type: type_name}, ...}
        :model_name: String that is used as new model's name.
        :raml_resource: Instance of pyraml.entities.RamlResource.
        :es_based: Boolean indicating if generated model should be a
            subclass of Elasticsearch-based document class or not.
            It True, ESBaseDocument is used; BaseDocument is used otherwise.
            Defaults to True.
        :predefined_fields: Dictionary of {field_name: field_obj} of fields
            that are already instantiated.
    """
    base_cls = engine.ESBaseDocument if es_based else engine.BaseDocument
    model_name = str(model_name)
    metaclass = type(base_cls)
    auth_model = schema.get('auth_model', False)
    if auth_model:
        bases = (AuthModelDefaultMixin, base_cls)
    else:
        bases = (base_cls,)
    attrs = {
        '__tablename__': model_name.lower(),
        '_public_fields': schema.get('public_fields') or [],
        '_auth_fields': schema.get('auth_fields') or [],
        '_nested_relationships': schema.get('nested_relationships') or [],
    }
    # Generate fields from properties
    properties = schema.get('properties', {})
    for field_name, props in properties.items():
        if field_name in attrs:
            continue

        field_kwargs = {
            'required': bool(props.get('required'))
        }
        field_kwargs.update(props.get('args', {}) or {})

        for proc_key in ('before_validation', 'after_validation'):
            processors = field_kwargs.get(proc_key, [])
            field_kwargs[proc_key] = [
                registry.get(name) for name in processors]

        raml_type = (props.get('type', 'string') or 'string').lower()
        if raml_type not in type_fields:
            raise ValueError('Unknown type: {}'.format(raml_type))

        field_cls = type_fields[raml_type]

        if field_cls is engine.Relationship:
            prepare_relationship(
                field_name, field_kwargs['document'], raml_resource)
        if field_cls is engine.ForeignKeyField:
            key = 'ref_column_type'
            field_kwargs[key] = type_fields[field_kwargs[key]]
        if field_cls is engine.ListField:
            key = 'item_type'
            field_kwargs[key] = type_fields[field_kwargs[key]]

        attrs[field_name] = field_cls(**field_kwargs)

    # Update model definition with methods and variables defined in registry
    attrs.update(registry.mget(model_name))

    # Generate new model class
    return metaclass(model_name, bases, attrs), auth_model



import pytest
from mock import Mock, patch

from ramses import utils


class TestUtils(object):

    def test_contenttypes(self):
        assert utils.ContentTypes.JSON == 'application/json'
        assert utils.ContentTypes.TEXT_XML == 'text/xml'
        assert utils.ContentTypes.MULTIPART_FORMDATA == \
            'multipart/form-data'
        assert utils.ContentTypes.FORM_URLENCODED == \
            'application/x-www-form-urlencoded'

    def test_convert_schema_json(self):
        schema = utils.convert_schema({'foo': 'bar'}, 'application/json')
        assert schema == {'foo': 'bar'}

    def test_convert_schema_json_error(self):
        with pytest.raises(TypeError) as ex:
            utils.convert_schema('foo', 'application/json')
        assert 'Schema is not a valid JSON' in str(ex.value)

    def test_convert_schema_xml(self):
        assert utils.convert_schema({'foo': 'bar'}, 'text/xml') is None

    def test_is_restful_uri(self):
        assert utils.is_restful_uri('/collection')
        assert utils.is_restful_uri('/{id}')
        assert not utils.is_restful_uri('/collection/{id}')

    def test_is_dynamic_uri(self):
        assert utils.is_dynamic_uri('/{id}')
        assert not utils.is_dynamic_uri('/collection')

    def test_clean_dynamic_uri(self):
        clean = utils.clean_dynamic_uri('/{item_id}')
        assert clean == 'item_id'

    def test_generate_model_name(self):
        model_name = utils.generate_model_name('/collectionitems')
        assert model_name == 'Collectionitem'

    def test_find_dynamic_resource(self):
        resource = Mock(resources={'/items': 'foo', '/{id}': 'bar'})
        assert utils.find_dynamic_resource(resource) == 'bar'

    def test_find_dynamic_resource_no_dynamic(self):
        resource = Mock(resources={'/items': 'foo'})
        assert utils.find_dynamic_resource(resource) is None

    def test_find_dynamic_resource_no_resources(self):
        resource = Mock(resources=None)
        assert utils.find_dynamic_resource(resource) is None

    def test_dynamic_part_name(self):
        resource = Mock(resources={'/items': 'foo', '/{myid}': 'bar'})
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_myid'

    def test_dynamic_part_name_no_dynamic(self):
        resource = Mock(resources={'/items': 'foo'})
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_default_id'

    def test_dynamic_part_name_no_resources(self):
        resource = Mock(resources=None)
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_default_id'

    def test_resource_view_attrs_no_dynamic_subres(self):
        resource = Mock(
            resources={
                '/items': 'foo'},
            methods={
                'get': '', 'post': '', 'put': '',
                'patch': '', 'delete': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set(['create', 'delete_many', 'index', 'update_many'])

    def test_resource_view_attrs_dynamic_subres(self):
        resource = Mock(
            resources={
                '/items': 'foo',
                '/{id}': Mock(methods={
                    'get': '', 'put': '', 'patch': '',
                    'DELETE': ''
                })
            },
            methods={
                'get': '', 'POST': '', 'put': '',
                'patch': '', 'delete': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set([
            'create', 'delete_many', 'index', 'update_many',
            'show', 'update', 'delete'
        ])

    def test_resource_view_attrs_singular(self):
        resource = Mock(
            resources={
                '/items': 'foo',
            },
            methods={
                'get': '', 'POST': '', 'put': '',
                'patch': '', 'delete': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=True)
        assert attrs == set(['create', 'delete', 'show', 'update'])

    def test_resource_view_attrs_no_subresources(self):
        resource = Mock(
            resources=None,
            methods={
                'get': '', 'POST': '', 'put': '',
                'patch': '', 'delete': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set(['create', 'delete_many', 'index', 'update_many'])

    def test_resource_view_attrs_no_methods(self):
        resource = Mock(
            resources={
                '/items': 'foo',
            },
            methods=None
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set()

    def test_resource_view_attrs_dynamic_subres_no_methods(self):
        resource = Mock(
            resources={
                '/items': 'foo',
                '/{id}': Mock(methods=None)
            },
            methods={
                'get': '', 'POST': '', 'put': '',
                'patch': '', 'delete': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set([
            'create', 'delete_many', 'index', 'update_many',
        ])

    def test_resource_view_attrs_not_supported_method(self):
        resource = Mock(
            resources={
                '/items': 'foo',
            },
            methods={'nice_method': ''}
        )
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set()

    def test_resource_schema(self):
        body = Mock(schema={'foo': 'bar'})
        method = Mock(body={'application/json': body})
        resource = Mock(methods={'post': method})
        assert utils.resource_schema(resource) == {'foo': 'bar'}
        resource = Mock(methods={'put': method})
        assert utils.resource_schema(resource) == {'foo': 'bar'}
        resource = Mock(methods={'patch': method})
        assert utils.resource_schema(resource) == {'foo': 'bar'}

    def test_resource_schema_methods_order(self):
        body = Mock(schema={'foo': 'bar'})
        method = Mock(body={'application/json': body})
        body2 = Mock(schema={'foo2': 'bar2'})
        method2 = Mock(body={'application/json': body2})
        resource = Mock(methods={
            'post': method,
            'patch': method2,
        })
        assert utils.resource_schema(resource) == {'foo': 'bar'}

    def test_resource_schema_no_propper_method(self):
        with pytest.raises(ValueError) as ex:
            utils.resource_schema(Mock(methods=None))
        assert str(ex.value) == 'No methods to setup database schema from'

        with pytest.raises(ValueError) as ex:
            utils.resource_schema(Mock(methods={'get': ''}))
        assert str(ex.value) == 'No methods to setup database schema from'

    def test_resource_schema_no_method_body(self):
        method = Mock(body=None)
        resource = Mock(methods={'post': method})
        assert utils.resource_schema(resource) is None

    def test_resource_schema_schema_none(self):
        body = Mock(schema=None)
        method = Mock(body={'application/json': body})
        resource = Mock(methods={'post': method})
        assert utils.resource_schema(resource) is None

    def test_resource_schema_invalid_content_type(self):
        body = Mock(schema={'foo': 'bar'})
        method = Mock(body={'dsadadasdasdasdasd': body})
        resource = Mock(methods={'post': method})
        assert utils.resource_schema(resource) is None

    def test_is_dynamic_resource_no_resource(self):
        assert not utils.is_dynamic_resource(None)

    def test_is_dynamic_resource_no_parent(self):
        resource = Mock(parentResource=None)
        assert not utils.is_dynamic_resource(resource)

    def test_is_dynamic_resource(self):
        resource = Mock()
        parent = Mock(resources={'/{id}': resource})
        resource.parentResource = parent
        assert utils.is_dynamic_resource(resource)

    def test_is_dynamic_resource_not_dynamic(self):
        resource = Mock()
        parent = Mock(resources={'/items': resource})
        resource.parentResource = parent
        assert not utils.is_dynamic_resource(resource)

    def test_is_dynamic_resource_no_resources(self):
        resource = Mock()
        parent = Mock(resources=None)
        resource.parentResource = parent
        assert not utils.is_dynamic_resource(resource)

    def test_get_static_parent(self):
        resource = Mock()
        parent = Mock(
            resources={'/{id}': resource},
            parentResource=Mock(resources=None))
        resource.parentResource = parent
        assert utils.get_static_parent(resource) is parent

    def test_get_static_parent_nested(self):
        resource = Mock()
        parent2 = Mock(
            parentResource=Mock(resources=None))
        parent = Mock(
            resources={'/{id}': resource},
            parentResource=parent2)
        parent2.resources = {'/{id}': parent}
        resource.parentResource = parent
        assert utils.get_static_parent(resource) is parent2

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_no_static_parent(self, mock_schema, mock_par):
        mock_par.return_value = None
        assert not utils.attr_subresource('foo', 1)
        mock_par.assert_called_once_with('foo')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_no_schema(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = None
        assert not utils.attr_subresource('foo', 1)
        mock_par.assert_called_once_with('foo')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_not_attr(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    'type': 'string'
                }
            }
        }
        assert not utils.attr_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_dict(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    'type': 'dict'
                },
                'route_name2': {
                    'type': 'list'
                }
            }
        }
        assert utils.attr_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource')
        mock_schema.assert_called_once_with(parent)
        assert utils.attr_subresource('resource', 'route_name2')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_no_static_parent(self, mock_schema, mock_par):
        mock_par.return_value = None
        assert not utils.singular_subresource('foo', 1)
        mock_par.assert_called_once_with('foo')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_no_schema(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = None
        assert not utils.singular_subresource('foo', 1)
        mock_par.assert_called_once_with('foo')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_not_attr(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    'type': 'string'
                }
            }
        }
        assert not utils.singular_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_dict(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    'type': 'relationship',
                    'args': {'uselist': False}
                },
            }
        }
        assert utils.singular_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource')
        mock_schema.assert_called_once_with(parent)

    def test_closest_secured_by(self):
        parent = Mock(securedBy=['foo'])
        resource = Mock(securedBy=['bar'], parentResource=parent)
        assert utils.closest_secured_by(resource) == ['bar']

    def test_closest_secured_by_parent(self):
        parent = Mock(securedBy=['foo'])
        resource = Mock(securedBy=None, parentResource=parent)
        assert utils.closest_secured_by(resource) == ['foo']

    def test_closest_secured_no_secured(self):
        parent = Mock(securedBy=None, parentResource=None)
        resource = Mock(securedBy=None, parentResource=parent)
        assert utils.closest_secured_by(resource) == []



import logging

from pyramid.security import (
    Allow, Deny,
    Everyone, Authenticated,
    ALL_PERMISSIONS)
from nefertari.acl import SelfParamMixin


from .views import collection_methods, item_methods
from . import registry


log = logging.getLogger(__name__)


actions = {
    'allow': Allow,
    'deny': Deny,
}
special_principals = {
    'everyone': Everyone,
    'authenticated': Authenticated,
}
ALLOW_ALL = (Allow, Everyone, ALL_PERMISSIONS)


def methods_to_perms(perms, methods_map):
    """ Convert permissions ("perms") which are either HTTP methods or
    the keyword 'all' into a set of valid Pyramid permissions.

    Arguments:
        :perms: List or comma-separated string of HTTP methods, or 'all'
        :methods_map: Map of HTTP methods to permission names (nefertari view
            methods)
    """
    if isinstance(perms, basestring):
        perms = perms.split(',')
    perms = [perm.strip().lower() for perm in perms]
    if 'all' in perms:
        return ALL_PERMISSIONS
    else:
        try:
            return [methods_map[p] for p in perms]
        except KeyError:
            raise ValueError(
                'Unknown method name in permissions: {}. Valid methods: '
                '{}'.format(perms, methods_map.keys()))


def parse_acl(acl_string, methods_map):
    """ Parse raw string :acl_string: of RAML-defined ACLs.

    If :acl_string: is blank or None, all permissions are given.
    Values of ACL action and principal are parsed using `actions` and
    `special_principals` maps and are looked up after `strip()` and `lower()`.

    ACEs in :acl_string: may be separated by newlines or semicolons.
    Action, principal and permission lists must be separated by spaces.
    Permissions must be comma-separated.
    E.g. 'allow everyone get,post,patch' and 'deny authenticated delete'

    Arguments:
        :acl_string: Raw RAML string containing defined ACEs.
        :methods_map: Map of HTTP methods to nefertari method handler names.
    """
    if not acl_string:
        return [ALLOW_ALL]

    aces_list = acl_string.replace('\n', ';').split(';')
    aces_list = [ace.strip().split(' ', 2) for ace in aces_list if ace]
    aces_list = [(a, b, c.split(',')) for a, b, c in aces_list]
    result_acl = []

    for action_str, princ_str, perms in aces_list:
        # Process action
        action_str = action_str.strip().lower()
        action = actions.get(action_str)
        if action is None:
            raise ValueError('Unknown ACL action: {}. Valid actions: {}'.format(
                action_str, actions.keys()))

        # Process principal
        princ_str = princ_str.strip().lower()
        if princ_str in special_principals:
            principal = special_principals[princ_str]
        elif princ_str.startswith('{{'):
            princ_str = princ_str.strip('{{').strip('}}').strip()
            principal = registry.get(princ_str)
        else:
            principal = 'g:' + princ_str

        # Process permissions
        permissions = methods_to_perms(perms, methods_map)

        result_acl.append((action, principal, permissions))

    return result_acl


class BaseACL(SelfParamMixin):
    """ ACL Base class. """
    __context_class__ = None
    collection_acl = None
    item_acl = None

    def __init__(self, request):
        super(BaseACL, self).__init__()
        self.request = request

    def _apply_callables(self, acl, methods_map, obj=None):
        """ Iterate over ACEs from :acl: and apply callable principals if any.

        Principals are passed 3 arguments on call:
            :ace: Single ACE object that looks like (action, callable,
                permission or [permission])
            :request: Current request object
            :obj: Object instance to be accessed via the ACL
        Principals must return a single ACE or a list of ACEs.

        Arguments:
            :acl: Sequence of valid Pyramid ACEs which will be processed
            :methods_map: Map of HTTP methods to nefertari view method names
                (permissions)
            :obj: Object to be accessed via the ACL
        """
        new_acl = []
        for i, ace in enumerate(acl):
            principal = ace[1]
            if callable(principal):
                ace = principal(ace=ace, request=self.request, obj=obj)
                if not ace:
                    continue
                if not isinstance(ace[0], (list, tuple)):
                    ace = [ace]
                ace = [(a, b, methods_to_perms(c, methods_map))
                       for a, b, c in ace]
            else:
                ace = [ace]
            new_acl += ace
        return new_acl

    def __acl__(self):
        """ Apply callables to `self.collection_acl` and return result. """
        return self._apply_callables(
            acl=self.collection_acl,
            methods_map=collection_methods)

    def context_acl(self, obj):
        """ Apply callables to `self.item_acl` and return result. """
        return self._apply_callables(
            acl=self.item_acl,
            methods_map=item_methods,
            obj=obj)

    def __getitem__(self, key):
        """ Get item using method depending on value of `self.es_based` """
        key = self.resolve_self_key(key)
        if self.es_based:
            return self.getitem_es(key=key)
        else:
            return self.getitem_db(key=key)

    def getitem_db(self, key):
        """ Get item with ID of :key: from database """
        pk_field = self.__context_class__.pk_field()
        obj = self.__context_class__.get_resource(
            **{pk_field: key})
        obj.__acl__ = self.context_acl(obj)
        obj.__parent__ = self
        obj.__name__ = key
        return obj

    def getitem_es(self, key):
        """ Get item with ID of :key: from elasticsearch """
        from nefertari.elasticsearch import ES
        es = ES(self.__context_class__.__name__)
        pk_field = self.__context_class__.pk_field()
        kwargs = {
            pk_field: key,
            '_limit': 1,
            '__raise_on_empty': True,
        }
        obj = es.get_collection(**kwargs)[0]
        obj.__acl__ = self.context_acl(obj)
        obj.__parent__ = self
        obj.__name__ = key
        return obj


def generate_acl(context_cls, raml_resource, parsed_raml, es_based=True):
    """ Generate an ACL.

    Generated ACL class has a `__context_class__` attribute set to :context_cls:.

    ACLs used for collection and item access control are generated from a
    security scheme which has a name of :raml_resource.securedBy[0]:.
    If :raml_resource: has no `securedBy` schemes defined then ALLOW_ALL ACL is
    used.
    If the `collection` or `item` settings are empty, then ALLOW_ALL ACL is used.

    Arguments:
        :context_cls: Generated model class
        :raml_resource: Instance of pyraml.entities.RamlResource for which
            ACL is being generated
        :parsed_raml: Whole parsed RAML object
        :es_based: Boolean inidicating whether ACL should query ES or not
            when getting an object
    """
    secured_by = raml_resource.securedBy or []

    if not secured_by:
        log.debug('No ACL scheme applied. Giving all permissions')
        collection_acl = item_acl = [ALLOW_ALL]
    else:
        log.debug('{} ACL scheme applied'.format(secured_by[0]))
        security_schemes = parsed_raml.securitySchemes or {}
        sec_scheme = security_schemes.get(secured_by[0])
        if sec_scheme is None:
            raise ValueError('Undefined ACL security scheme: {}'.format(
                secured_by[0]))
        settings = sec_scheme.settings or {}
        collection_acl = parse_acl(
            acl_string=settings.get('collection'),
            methods_map=collection_methods)
        item_acl = parse_acl(
            acl_string=settings.get('item'),
            methods_map=item_methods)

    class GeneratedACL(BaseACL):
        __context_class__ = context_cls

        def __init__(self, request, es_based=es_based):
            super(GeneratedACL, self).__init__(request=request)
            self.es_based = es_based
            self.collection_acl = collection_acl
            self.item_acl = item_acl

    return GeneratedACL



"""
Auth module that contains all code needed for authentication/authorization
policies setup.

In particular:
    :includeme: Function that actually creates routes listed above and
        connects view to them
    :create_admin_user: Function that creates system/admin user
    :_setup_ticket_policy: Setup Pyramid AuthTktAuthenticationPolicy
    :_setup_apikey_policy: Setup nefertari.ApiKeyAuthenticationPolicy
    :setup_auth_policies: Runs generation of particular auth policy
"""
import logging

import transaction
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy

from nefertari.utils import dictset
from nefertari.json_httpexceptions import *
from nefertari.authentication.policies import ApiKeyAuthenticationPolicy
from nefertari.authentication.views import (
    TicketAuthenticationView, TokenAuthenticationView)

log = logging.getLogger(__name__)


def _setup_ticket_policy(config, params):
    """ Setup Pyramid AuthTktAuthenticationPolicy.

    Notes:
      * Initial `secret` params value is considered to be a name of config
        param that represents a cookie name.
      * `auth_model.get_groups_by_userid` is used as a `callback`.
      * Special processing is applied to boolean params to convert string
        values like 'True', 'true' to booleans. This is done because pyraml
        parser currently does not support setting value being a boolean.
      * Also connects basic routes to perform authentication actions.

    Arguments:
        :config: Pyramid Configurator instance.
        :params: Nefertari dictset which contains security scheme `settings`.
    """
    log.info('Configuring Pyramid Ticket Authn policy')
    if 'secret' not in params:
        raise ValueError(
            'Missing required security scheme settings: secret')
    bool_keys = ('secure', 'include_ip', 'http_only', 'wild_domain', 'debug',
                 'parent_domain')
    for key in bool_keys:
        if key in params:
            params[key] = params.asbool(key)

    params['secret'] = config.registry.settings[params['secret']]

    auth_model = config.registry.auth_model
    params['callback'] = auth_model.get_groups_by_userid

    config.add_request_method(
        auth_model.get_authuser_by_userid, 'user', reify=True)

    policy = AuthTktAuthenticationPolicy(**params)

    class RamsesTicketAuthenticationView(TicketAuthenticationView):
        _model_class = config.registry.auth_model

    config.add_route('login', '/auth/login')
    config.add_view(
        view=RamsesTicketAuthenticationView,
        route_name='login', attr='login', request_method='POST')

    config.add_route('logout', '/auth/logout')
    config.add_view(
        view=RamsesTicketAuthenticationView,
        route_name='logout', attr='logout')

    config.add_route('register', '/auth/register')
    config.add_view(
        view=RamsesTicketAuthenticationView,
        route_name='register', attr='register', request_method='POST')

    return policy


def _setup_apikey_policy(config, params):
    """ Setup `nefertari.ApiKeyAuthenticationPolicy`.

    Notes:
      * User may provide model name in :params['user_model']: do define
        the name of the user model.
      * `auth_model.get_groups_by_token` is used to perform username & token
        check
      * `auth_model.get_token_credentials` is used to get username and token
        from userid
      * Also connects basic routes to perform authentication actions.

    Arguments:
        :config: Pyramid Configurator instance.
        :params: Nefertari dictset which contains security scheme `settings`.
    """
    log.info('Configuring ApiKey Authn policy')

    auth_model = config.registry.auth_model
    params['check'] = auth_model.get_groups_by_token
    params['credentials_callback'] = auth_model.get_token_credentials
    params['user_model'] = auth_model
    config.add_request_method(
        auth_model.get_authuser_by_name, 'user', reify=True)

    policy = ApiKeyAuthenticationPolicy(**params)

    class RamsesTokenAuthenticationView(TokenAuthenticationView):
        _model_class = auth_model

    config.add_route('token', '/auth/token')
    config.add_view(
        view=RamsesTokenAuthenticationView,
        route_name='token', attr='claim_token', request_method='POST')

    config.add_route('reset_token', '/auth/reset_token')
    config.add_view(
        view=RamsesTokenAuthenticationView,
        route_name='reset_token', attr='reset_token', request_method='POST')

    config.add_route('register', '/auth/register')
    config.add_view(
        view=RamsesTokenAuthenticationView,
        route_name='register', attr='register', request_method='POST')

    return policy


""" Map of `security_scheme_type`: `generator_function`, where:

  * `security_scheme_type`: String that represents RAML security scheme type
    name that should be used to apply a particular authentication system.
  * `generator_function`: Function that receives instance of Pyramid
    Configurator instance and dictset of security scheme settings and returns
    generated Pyramid authentication policy instance.

"""
AUTHENTICATION_POLICIES = {
    'x-ApiKey': _setup_apikey_policy,
    'x-Ticket': _setup_ticket_policy,
}


def setup_auth_policies(config, raml_data):
    """ Setup authentication, authorization policies.

    Performs basic validation to check all the required values are present
    and performs authentication, authorization policies generation using
    generator functions from `AUTHENTICATION_POLICIES`.

    Arguments:
        :config: Pyramid Configurator instance.
        :raml_data: Instance of pyraml.parser.entities.RamlRoot.
    """
    log.info('Configuring auth policies')
    secured_by = filter(bool, (raml_data.securedBy or []))
    if not secured_by:
        log.info('API is not secured. `securedBy` attribute value missing.')
        return
    secured_by = secured_by[0]

    if secured_by not in raml_data.securitySchemes:
        raise ValueError(
            'Not defined security scheme used in `securedBy`: {}'.format(
                secured_by))

    scheme = raml_data.securitySchemes[secured_by]
    if scheme.type not in AUTHENTICATION_POLICIES:
        raise ValueError('Not supported security scheme type: {}'.format(
            scheme.type))

    # Setup Authentication policy
    policy_generator = AUTHENTICATION_POLICIES[scheme.type]
    params = dictset(scheme.settings or {})
    authn_policy = policy_generator(config, params)
    config.set_authentication_policy(authn_policy)

    # Setup Authorization policy
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)


def create_admin_user(config):
    log.info('Creating system user')
    settings = config.registry.settings
    try:
        auth_model = config.registry.auth_model
        s_user = settings['system.user']
        s_pass = settings['system.password']
        s_email = settings['system.email']
        user, created = auth_model.get_or_create(
            username=s_user,
            defaults=dict(
                password=s_pass,
                email=s_email,
                groups=['admin']
            ))
        if created:
            transaction.commit()
    except KeyError as e:
        log.error('Failed to create system user. Missing config: %s' % e)


def includeme(config):
    log.info('Creating admin user')
    create_admin_user(config)



import logging

import inflection


log = logging.getLogger(__name__)


class ContentTypes(object):
    """ ContentType values.

    """
    JSON = 'application/json'
    TEXT_XML = 'text/xml'
    MULTIPART_FORMDATA = 'multipart/form-data'
    FORM_URLENCODED = 'application/x-www-form-urlencoded'


def convert_schema(raml_schema, schema_ct):
    """ Restructure `raml_schema` to a dictionary that has 'properties'
    as well as other schema keys/values.

    The resulting dictionary looks like this::

    {
        "properties": {
            "field1": {
                "required": boolean,
                "type": ...,
                ...more field options
            },
            ...more properties
        },
        "public_fields": [...],
        "auth_fields": [...],
        ...more schema options
    }

    Arguments:
        :raml_schema: pyraml.entities.RamlBody.schema.
        :schema_ct: ContentType of the schema as a string from RAML file. Only
            JSON is currently supported.
    """
    if schema_ct == ContentTypes.JSON:
        if not isinstance(raml_schema, dict):
            raise TypeError(
                'Schema is not a valid JSON. Please check your '
                'schema syntax.\n{}...'.format(str(raml_schema)[:60]))
        return raml_schema
    if schema_ct == ContentTypes.TEXT_XML:
        # Process XML schema
        pass


def is_restful_uri(uri):
    """ Check whether `uri` is a RESTful uri.

    Uri is assumed to be restful if it only contains a single token.
    E.g. 'stories' and 'users' but NOT 'stories/comments' and 'users/{id}'

    Arguments:
        :uri: URI as a string
    """
    uri = uri.strip('/')
    return '/' not in uri


def is_dynamic_uri(uri):
    """ Determine whether `uri` is a dynamic uri or not.

    Assumes a dynamic uri is one that ends with '}' which is a Pyramid
    way to define dynamic parts in uri.

    Arguments:
        :uri: URI as a string.
    """
    return uri.strip('/').endswith('}')


def clean_dynamic_uri(uri):
    """ Strips /, {, } from dynamic `uri`.

    Arguments:
        :uri: URI as a string.
    """
    return uri.replace('/', '').replace('{', '').replace('}', '')


def generate_model_name(name):
    """ Generate model name.

    Arguments:
        :name: String representing a field or route name.
    """
    model_name = inflection.camelize(name.strip('/'))
    return inflection.singularize(model_name)


def find_dynamic_resource(raml_resource):
    """ Find dymanic resource in :raml_resource:'s resources.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource.
    """
    subresources = raml_resource.resources or {}
    dynamic_resources = [res for uri, res in subresources.items()
                         if is_dynamic_uri(uri)]
    return dynamic_resources[0] if dynamic_resources else None


def dynamic_part_name(raml_resource, clean_uri, pk_field):
    """ Generate a dynamic part for a resource :raml_resource:.

    A dynamic part is generated using 2 parts: :clean_uri: of the resource and
    the dynamic part of any dymanic subresources. If :raml_resource: has no dynamic
    subresources, 'id' is used as the 2nd part.
    E.g. if your dynamic part on route 'stories' is named 'superId' then dynamic
    part will be 'stories_superId'.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource for which
            dynamic part name is being generated.
        :clean_uri: Cleaned URI of :raml_resource:
    """
    subresources = raml_resource.resources or {}
    dynamic_uris = [uri for uri in subresources.keys() if is_dynamic_uri(uri)]
    if dynamic_uris:
        dynamic_part = clean_dynamic_uri(dynamic_uris[0])
    else:
        dynamic_part = pk_field
    return '_'.join([clean_uri, dynamic_part])


def resource_view_attrs(raml_resource, singular=False):
    """ Generate view method names needed for `raml_resource` view.

    Collects HTTP method names from `raml_resource.methods` and
    dynamic child `methods` if a child exists. Collected methods are
    then translated  to `nefertari.view.BaseView` method names,
    each of which is used to process a particular HTTP method request.

    Maps of {HTTP_method: view_method} `collection_methods` and
    `item_methods` are used to convert collection and item methods
    respectively.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource.
    """
    from .views import collection_methods, item_methods

    # Singular resource doesn't have collection methods though
    # it looks like a collection
    if singular:
        collection_methods = item_methods

    http_methods = (raml_resource.methods or {}).keys()
    attrs = [collection_methods.get(m.lower()) for m in http_methods]

    # Check if resource has dynamic subresource like collection/{id}
    subresources = raml_resource.resources or {}
    dynamic_res = [res for uri, res in subresources.items()
                   if is_dynamic_uri(uri)]

    # If dynamic subresource exists, add its methods to attrs, as both
    # resources are handled by a single view
    if dynamic_res and dynamic_res[0].methods:
        http_submethods = (dynamic_res[0].methods or {}).keys()
        attrs += [item_methods.get(m.lower()) for m in http_submethods]

    return set(filter(bool, attrs))


def resource_schema(raml_resource):
    """ Get schema properties of RAML resource :raml_resource:.

    The process follows these steps:
      * :raml_resource: post, put, patch methods body schemas are checked
        to see if a schema is defined.
      * If found, the schema is restructured into a dictionary of form
        {field_name: {required: boolean, type: type_name}} and returned.

    Arguments:
        :raml_resource: Instance of pyraml.entities.RamlResource.
    """
    log.info('Searching for model schema')
    schemas = (ContentTypes.JSON, ContentTypes.TEXT_XML)
    methods = raml_resource.methods or {}

    # Get 'schema' from particular methods' bodies
    method = (methods.get('post') or
              methods.get('put') or
              methods.get('patch'))
    if not method:
        raise ValueError('No methods to setup database schema from')

    # Find what schema from 'schemas' is defined
    body = method.body or {}
    for schema_ct in schemas:
        if schema_ct not in body:
            continue
        schema = body[schema_ct].schema
        if schema:
            return convert_schema(schema, schema_ct)
    log.debug('No model schema found.')


def is_dynamic_resource(raml_resource):
    """ Determine if :raml_resource: is a dynamic resource.

    Arguments:
        :raml_resource:Instance of pyraml.entities.RamlResource.
    """
    if not (raml_resource and raml_resource.parentResource):
        return False
    dyn = find_dynamic_resource(raml_resource.parentResource)
    return dyn is raml_resource


def get_static_parent(raml_resource):
    """ Get static parent resource of :raml_resource:.

    Arguments:
        :raml_resource:Instance of pyraml.entities.RamlResource.
    """
    parent = raml_resource.parentResource
    while is_dynamic_resource(parent):
        parent = parent.parentResource
    return parent


def attr_subresource(raml_resource, route_name):
    """ Determine if :raml_resource: is an attribute subresource.

    Attribute:
        :raml_resource: Instance of pyraml.entities.RamlResource.
        :route_name: Name of the :raml_resource:.
    """
    static_parent = get_static_parent(raml_resource)
    if static_parent is None:
        return False
    schema = resource_schema(static_parent) or {}
    properties = schema.get('properties', {})
    return (route_name in properties and
            properties[route_name]['type'] in ('dict', 'list'))


def singular_subresource(raml_resource, route_name):
    """ Determine if :raml_resource: is a singular subresource.

    Attribute:
        :raml_resource: Instance of pyraml.entities.RamlResource.
        :route_name: Name of the :raml_resource:.
    """
    static_parent = get_static_parent(raml_resource)
    if static_parent is None:
        return False
    schema = resource_schema(static_parent) or {}
    properties = schema.get('properties', {})
    if route_name not in properties:
        return False
    is_obj = properties[route_name]['type'] == 'relationship'
    args = properties[route_name].get('args', {})
    single_obj = not args.get('uselist', True)
    return is_obj and single_obj


def closest_secured_by(raml_resource):
    """ Get closest securedBy attr valid for current resource.

    Finds the attr by going up the inheritance tree and stops
    when first 'securedBy' attr is met.

    Attributes:
        :raml_resource: Instance of pyraml.entities.RamlResource.
    """
    secured_by = []
    resource = raml_resource

    while not secured_by and resource:
        secured_by = resource.securedBy or []
        resource = resource.parentResource

    return secured_by



from pyramid.scaffolds import PyramidTemplate
import binascii
import os
from os import urandom, chdir
import subprocess


class RamsesStarterTemplate(PyramidTemplate):
    _template_dir = 'ramses_starter'
    summary = 'Ramses starter'

    def pre(self, command, output_dir, vars):
        dbengine_choices = {'1':'sqla', '2':'mongodb'}
        vars['engine'] = dbengine_choices[raw_input("""
        Which DB backend would you like to use: 1)'sqla' or 2)'mongodb'?:
        """)]
        vars['random_string'] = binascii.hexlify(os.urandom(20))
        if vars['package'] == 'site':
            raise ValueError("""
                "Site" is a reserved keyword in Python.
                 Please use a different project name. """)

    def post(self, command, output_dir, vars):
        os.chdir(str(output_dir))
        subprocess.call('pip install -r requirements.txt', shell=True)
        subprocess.call('pip install nefertari-{}'.format(vars['engine']), shell=True)
        msg = """Goodbye boilerplate! Welcome to Ramses."""
        self.out(msg)




import logging

import pyraml.parser
from nefertari.acl import RootACL as NefertariRootACL
from nefertari.utils import dictset

from .generators import generate_server, generate_models


log = logging.getLogger(__name__)


def includeme(config):
    Settings = dictset(config.registry.settings)
    config.include('nefertari.engine')
    config.include('nefertari')
    config.include('nefertari.view')
    config.include('nefertari.elasticsearch')

    # Process nefertari settings
    if Settings.asbool('debug'):
        log.warning('*** DEBUG DEBUG DEBUG mode ***')
        config.add_tween('nefertari.tweens.get_tunneling')

    if Settings.asbool('cors.enable'):
        config.add_tween('nefertari.tweens.cors')

    if Settings.asbool('ssl_middleware.enable'):
        config.add_tween('nefertari.tweens.ssl')

    if Settings.asbool('request_timing.enable'):
        config.add_tween('nefertari.tweens.request_timing')

    # Set root factory
    config.root_factory = NefertariRootACL

    # Process auth settings
    root = config.get_root_resource()
    ramses_auth = Settings.asbool('ramses.auth', False)
    root.auth = ramses_auth

    log.info('Parsing RAML')
    parsed_raml = pyraml.parser.load(Settings['ramses.raml_schema'])

    log.info('Starting models generation')
    generate_models(config, raml_resources=parsed_raml.resources)

    if ramses_auth:
        if getattr(config.registry, 'auth_model', None) is None:
            from nefertari.authentication.models import AuthUser
            config.registry.auth_model = AuthUser
        from .auth import setup_auth_policies
        setup_auth_policies(config, parsed_raml)

    log.info('Starting server generation')
    generate_server(parsed_raml, config)

    log.info('Running nefertari.engine.setup_database')
    from nefertari.engine import setup_database
    setup_database(config)

    from nefertari.elasticsearch import ES
    ES.setup_mappings()

    if ramses_auth:
        config.include('ramses.auth')

    log.info('Server succesfully generated\n')



import pytest
from mock import Mock, patch, call

from .fixtures import engine_mock


class TestHelperFunctions(object):

    @patch('ramses.models.engine')
    def test_get_existing_model_not_found(self, mock_eng):
        from ramses import models
        mock_eng.get_document_cls.side_effect = ValueError
        model_cls = models.get_existing_model('Foo')
        assert model_cls is None
        mock_eng.get_document_cls.assert_called_once_with('Foo')

    @patch('ramses.models.engine')
    def test_get_existing_model_found(self, mock_eng):
        from ramses import models
        mock_eng.get_document_cls.return_value = 1
        model_cls = models.get_existing_model('Foo')
        assert model_cls == 1
        mock_eng.get_document_cls.assert_called_once_with('Foo')

    @patch('ramses.models.find_dynamic_resource')
    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship_model_exists(self, mock_get, mock_find):
        from ramses import models
        models.prepare_relationship('foobar', 'Story', 'raml_resource')
        mock_get.assert_called_once_with('Story')
        assert not mock_find.called

    @patch('ramses.models.find_dynamic_resource')
    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship_no_subresource(self, mock_get, mock_find):
        from ramses import models
        mock_get.return_value = None
        mock_find.return_value = Mock(resources={'/foo': 'bar'})
        with pytest.raises(ValueError) as ex:
            models.prepare_relationship('foobar', 'Story', 'raml_resource')
        expected = ('Model `Story` used in relationship `foobar` '
                    'is not defined')
        assert str(ex.value) == expected
        mock_get.assert_called_once_with('Story')
        mock_find.assert_called_once_with('raml_resource')

    @patch('ramses.generators.setup_data_model')
    @patch('ramses.models.find_dynamic_resource')
    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship(self, mock_get, mock_find, mock_setup):
        from ramses import models
        mock_get.return_value = None
        mock_find.return_value = Mock(
            resources={'/foobar': 'stories_resource'})
        models.prepare_relationship('foobar', 'Story', 'raml_resource')
        mock_get.assert_called_once_with('Story')
        mock_find.assert_called_once_with('raml_resource')
        mock_setup.assert_called_once_with('stories_resource', 'Story')


@patch('ramses.models.registry')
@pytest.mark.usefixtures('engine_mock')
class TestGenerateModelCls(object):

    def _test_schema(self):
        return {
            'properties': {},
            'auth_model': False,
            'public_fields': ['public_field1'],
            'auth_fields': ['auth_field1'],
            'nested_relationships': ['nested_field1']
        }

    def test_simple_case(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "required": True,
            "type": "float",
            "args": {
                "default": 0,
                "before_validation": ["zoo"],
                "after_validation": ["foo"]
            }
        }
        mock_reg.get.return_value = 1
        mock_reg.mget.return_value = {'foo': 'bar'}
        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None)
        assert not auth_model
        assert model_cls.__name__ == 'Story'
        assert hasattr(model_cls, 'progress')
        assert model_cls.__tablename__ == 'story'
        assert model_cls._public_fields == ['public_field1']
        assert model_cls._auth_fields == ['auth_field1']
        assert model_cls._nested_relationships == ['nested_field1']
        assert model_cls.foo == 'bar'
        assert issubclass(model_cls, models.engine.ESBaseDocument)
        assert not issubclass(model_cls, models.AuthModelDefaultMixin)
        models.engine.FloatField.assert_called_once_with(
            default=0, required=True, before_validation=[1],
            after_validation=[1])
        mock_reg.get.assert_has_calls([call('zoo'), call('foo')])
        mock_reg.mget.assert_called_once_with('Story')

    def test_auth_model(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {}
        schema['auth_model'] = True
        mock_reg.mget.return_value = {'foo': 'bar'}

        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None)
        assert auth_model
        assert issubclass(model_cls, models.AuthModelDefaultMixin)

    def test_db_based_model(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {}
        mock_reg.mget.return_value = {'foo': 'bar'}

        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None,
            es_based=False)
        assert issubclass(model_cls, models.engine.BaseDocument)
        assert not issubclass(model_cls, models.engine.ESBaseDocument)
        assert not issubclass(model_cls, models.AuthModelDefaultMixin)

    def test_unknown_field_type(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {'type': 'foobar'}
        mock_reg.mget.return_value = {'foo': 'bar'}

        with pytest.raises(ValueError) as ex:
            models.generate_model_cls(
                schema=schema, model_name='Story', raml_resource=None)
        assert str(ex.value) == 'Unknown type: foobar'

    @patch('ramses.models.prepare_relationship')
    def test_relationship_field(self, mock_prep, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            'type': 'relationship',
            'args': {'document': 'FooBar'}
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        mock_prep.assert_called_once_with('progress', 'FooBar', 1)

    def test_foreignkey_field(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "type": "foreign_key",
            "args": {
                "ref_column_type": "string"
            }
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        models.engine.ForeignKeyField.assert_called_once_with(
            required=False, ref_column_type=models.engine.StringField,
            before_validation=[], after_validation=[])

    def test_list_field(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "type": "list",
            "args": {
                "item_type": "integer"
            }
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        models.engine.ListField.assert_called_once_with(
            required=False, item_type=models.engine.IntegerField,
            before_validation=[], after_validation=[])

    def test_duplicate_field_name(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['_public_fields'] = {'type': 'interval'}
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        assert not models.engine.IntervalField.called
